function d = fastmarching(M, x0)
% M = load_off('../../Shapes/KIDS/0001.isometry.2.off');
% x0 = 1;

n = size(M.VERT,1);
m = size(M.TRIV,1);

%I = incidenceMatrix(M.TRIV);

% calculate length of all edges + adjacency list
mesh = cell(n,1); % mesh{i} is a 2xk matrix where k is number of neighbors of vertex i
mesh(:) = {[]};

for k=1:m
    indices = [M.TRIV(k,:), M.TRIV(k,1)];
    for i=1:3
       i1 = indices(i);
       i2 = indices(i+1);
       length = sqrt(sum((M.VERT(i1,:) - M.VERT(i2,:)).^2));
       mesh{i1} = [mesh{i1}, [i2; length; k], [i1; length; k]];
    end
end

% Initializing
R = MinHeap(n);
R.push(0,x0);
d = inf(n,1);
for i=1:n
    R.push(Inf,i);
end

% run fast marching

while (~R.isEmpty())
    [~, current] = R.pop();
    x1 = current(2);
    energy = current(1);
    N = mesh{x1};
    % dijkstra step
    for i=1:size(N,2)
        R.decrease(N(1,i), energy+N(2,i));
    end
    % fast  marching step
    for i=1:size(N,2)
       if(R.peakKey(N(1,i)) > -1)
          k = N(3,i);
          x2 = N(1,i);
          x3 = M.TRIV(k, M.TRIV(k,:) ~= x1 & M.TRIV(k,:) ~= x2);
          if R.peakKey(x2) < 0
              energy2 = d(x2);
          else
              energy2 = R.peakKey(x2);
          end
          if R.peakKey(x3) < 0
              energy3 = d(x3);
          else
              energy3 = R.peakKey(x3);
          end
          d3 = march( M.VERT([x1, x2, x3], :), [energy; energy2; energy3] ); 
          R.decrease(x3, d3);
       end
    end
    d(x1) = energy;
end

end