function d3 = march(X, d)
% MARCH Performes the fast marching update step. X is a 3x3 matrix
% containing coordinates of one vertex of the triangle in each column.
% The last column contains the the vertex that is supposed to be updated.
% d is a 3-vector containing the current distances of each vertex (in the
% same order as they are in X). 

X = X-repmat(X(:,3),1,3); % Move x3 to origin
X = X(:, 1:2);

if d(3) < 0
   d3 = Inf;
else
   d3 = d(3); 
end
d = d(1:2);

U = X(:,1)/norm(X(:,1));
V = cross(U,X(:,2))/norm(cross(U,X(:,2)));
W = cross(U,V);
R = [U,W,V]';
X = R*X;
V = X(1:2, 1:2);

Q = inv(V'*V);

a = sum(Q(:));
b = -2*sum(Q)*d;
c = d'*Q*d-1;
% a*d3^2+b*d3+c = 0
% d3 is the larger solution

p = (sqrt(b^2-4*a*c)-b)/(2*a);

n = inv(V')*(d - p * ones(2,1));
if Q*V'*n < 0
    d3 = min(d3,p);
else
    d3 = min([d3, d(1) + norm(V(:,1)), d(2) + norm(V(:,2))]);
end

end